import math
import pandas as pd
import numpy as np
import random
from simanneal import Annealer
import random
import numpy as np
import re
import warnings
warnings.simplefilter('ignore', FutureWarning)

def standard_g(n, size):
        #Strategy: Assign groups of size g or g+1, if that fails reduce group size by 1 and try again

        # Begin by trying to maximise the number of groups of size g i.e. a in a*g + b*(g+1) = n
        a = math.floor(n/size)

        # Then decrease a until a sufficient b is found
        b = 0
        # If a goes negative try a smaller group size (won't affect large n, small size)
        while(a >= 0):
            if (a*size + b*(size+1)) == n:
                return [size,a,b]
            elif (a*size + b*(size+1)) > n:
                a = a - 1
                b = 0
            else:
                b = b + 1
        else:
            return standard_g(n, size-1)

def results_to_csv(state, data, group_size, name='Allocation.csv', g=standard_g):
    grouping = np.array(g(len(state), group_size))
    number_of_groups = sum(grouping[1:])
    
    df = pd.DataFrame(columns=data.columns)

    index = 0
    for i in range(number_of_groups):

        size_of_group = grouping[0] if i < grouping[1] else grouping[0] + 1

        group_ids = state[index:(index+size_of_group)]

        index = index + size_of_group

        # Subtract one from each
        group_ids = [x-1 for x in group_ids]
        
        # Retrieve from data
        y = data.iloc[group_ids, :]
        df = df.append(y)
        df = df.append(pd.Series(), ignore_index=True)

    df.to_csv(name, index=False)

def email_to_department(email):
    email = email[email.index("@")+1:]
    email = email[0:email.index(".")]
    return email.lower()

def generate_data(n, m):
    # n is number of people
    # m is timeslots
    data = np.arange(n*m).reshape(n,m)
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            data[i,j] = random.randint(0,1)
    return data

def pad(x):
    return str(x) + " "

def test(x):
    return x+1

    
def get_color(x):
    if (x == 0):
        return "\033[1;41;41m 0"
    else:
        return "\033[1;42;42m 1"

def print_pretty_allocations(state, times, group_size, g = standard_g, extras = None):
    x = np.asarray(times)
    size = group_size
    digits = math.floor(round(math.log(len(state)-1,10),5)) + 1
    grouping = np.array(g(len(state), size))
    number_of_groups = sum(grouping[1:])

    index = 0
    for i in range(number_of_groups):

        size_of_group = grouping[0] if i < grouping[1] else grouping[0] + 1

        group_ids = state[index:(index+size_of_group)]

        index = index + size_of_group
        
        # Subtract one from each
        group_ids = [x-1 for x in group_ids]
        
        # Retrieve from x
        y = x[group_ids, :]
        
        if extras is not None:
            extra = extras.iloc[group_ids, :]
        
        print()
        
        for j in range(y.shape[0]):

            # Pad string for pretty formatting
            s = str(group_ids[j]+1)
            for i in range(int(digits - math.floor(math.log(group_ids[j]+1 if group_ids[j]+1 > 0 else 1,10)))-1):
                s = pad(s)
            
            
            # Add colours to make pretty
            for c in range(y[j,:].shape[0]):
                s = s + get_color(y[j,c])
            s = s + "\033[0m"
            
            if extras is not None:
                # Add extras
                for c in range(extra.iloc[j].shape[0]):
                    s = s + " ---- " + str(extra.iloc[j,c])
            
            print(s)



import math
import numpy as np
import random
from simanneal import Annealer
import random
import numpy as np
import lib.utils as utils

class SAAT(Annealer):

    def __init__(self, 
                 times, 
                 group_size, 
                 times_wt = 1, 
                 organise_wt = 0.5, 
                 time_in_company_wt = 0.5,
                 department_wt = 0.25,
                 no_time_penalty_wt = 0.5,
                 organise = None, 
                 department = None, 
                 time_in_company = None, 
                 g = None):
        
        if no_time_penalty_wt < 0 or no_time_penalty_wt > 1:
            raise Exception("no_time_penalty_weight must be between 0 and 1")
        
        state = list(range(1,times.shape[0]+1))
        self.x = np.asarray(times)
        self.size = group_size
        self.g = utils.standard_g if g is None else g
        self.organise = None if organise is None else np.asarray(organise)
        self.department = None if department is None else np.asarray(department)
        self.time_in_company = None if time_in_company is None else np.asarray(time_in_company)
        self.times_wt = times_wt*group_size
        self.organise_wt = organise_wt
        self.time_in_company_wt = time_in_company_wt
        self.department_wt = department_wt
        self.no_time_penalty_wt = no_time_penalty_wt
        self.copy_strategy = "slice"
        super(SAAT, self).__init__(state)
    
    def move(self):
        a = random.randint(0, len(self.state) - 1)
        b = random.randint(0, len(self.state) - 1)
        self.state[a], self.state[b] = self.state[b], self.state[a]

    def energy(self, logging=False):
        grouping = np.array(self.g(len(self.state), self.size))
        number_of_groups = sum(grouping[1:])
        
        index = 0
        total_score = 0
        
        if logging:
            print('Numer of groups: {}'.format(number_of_groups))
        
        for i in range(number_of_groups):
            
            local_score = 0
            
            if logging:
                print('Group: {}'.format(i))

            size_of_group = grouping[0] if i < grouping[1] else grouping[0] + 1

            group_ids = self.state[index:(index+size_of_group)]

            index = index + size_of_group

            # Subtract one from each
            group_ids = [x-1 for x in group_ids]
            
            # Retrieve from x
            y = self.x[group_ids, :]

            # Count number of matching columns
            count = 0
            for c in range(y.shape[1]):
                if sum(y[:,c]) == y.shape[0]:
                    count = count + 1
            
            # Count total free slots
            slots_count = sum(sum(y))
            
            # Summation
            score = count/slots_count
            local_score = local_score + score*self.times_wt
            
            if logging:
                print('Current total score after free slots: {}'.format(total_score))
                print('Score: {}, Slots: {}, Count: {}'.format(score, slots_count, count)) 
            
            # Retrieve from organise
            if self.organise is not None:
                y = self.organise[group_ids]

                if ((1 in y) or ('Yes' in y)):
                    local_score = local_score + 1*self.organise_wt
                elif ((0.5 in y) or ('If needed' in y)):
                    local_score = local_score + 0.5*self.organise_wt
                
            if logging:
                print('Current total score after organise: {}'.format(total_score))
                
            # Retrieve from time_in_company
            if self.time_in_company is not None:
                local_score = local_score + self.time_in_company_wt * len(np.unique(self.time_in_company[group_ids])) / len(self.time_in_company[group_ids])
            
            if logging:
                print('Current total score after time in org: {}'.format(total_score))
            
            # Retrieve from department
            if self.department is not None:
                local_score = local_score + self.department_wt * len(np.unique(self.department[group_ids])) / len(self.department[group_ids])
                
            if logging:
                print('Current total score after department: {}'.format(total_score))
                
            if logging:
                utils.print_pretty_allocations(self.state, self.x, self.size)
                
            # If no free slots, give penalty
            if count == 0:
                local_score = local_score*(1-self.no_time_penalty_wt)
                
            total_score = total_score + local_score
        
        return -total_score